'$DYNAMIC
DECLARE SUB fillTable (ATable() AS ANY)
DECLARE SUB extract (AInFile AS STRING, AOutFile AS STRING, AOffset AS LONG, ASize AS LONG)
DECLARE SUB checkFile (AFilename AS STRING)
DECLARE FUNCTION fileExists% (AFilename AS STRING)
DECLARE FUNCTION search% (AKey AS INTEGER, AArray() AS INTEGER)
DECLARE FUNCTION hash% (AFilename AS STRING)

CONST FALSE = 0
CONST TRUE = NOT FALSE
CONST ARCHIVEFILE = "GODDESS.LBR"

TYPE tableType
  offset AS LONG
  filesize AS LONG
  filename AS STRING * 12
END TYPE

PRINT "Vinyl Goddess From Mars extractor by Frenkel, 1-Sep-2006."

checkFile ARCHIVEFILE

DIM table(0 TO 0) AS tableType
fillTable table()

DIM fdummy AS STRING 'filename dummy
DIM kdummy AS STRING 'key dummy
FOR i% = LBOUND(table) TO UBOUND(table)
  fdummy = RTRIM$(table(i%).filename)
  IF fileExists%(fdummy) THEN
    'ask whether to overwrite or not
    PRINT "File " + fdummy + " already exists."
    PRINT "Do you want to overwrite " + fdummy + "? [y/N]"
    kdummy = ""
    DO WHILE kdummy = ""
      kdummy = INKEY$
    LOOP
    IF UCASE$(kdummy) = "Y" THEN
      'yes => overwrite
      KILL fdummy
      extract ARCHIVEFILE, fdummy, table(i%).offset, table(i%).filesize
    'ELSE
      'no => skip
    END IF
  ELSE
    'write
    extract ARCHIVEFILE, fdummy, table(i%).offset, table(i%).filesize
  END IF
NEXT i%
END

REM $STATIC
SUB checkFile (AFilename AS STRING)

IF NOT fileExists(AFilename) THEN
  PRINT "File " + AFilename + " does not exist."
  PRINT "Press any key to continue . . . "
  DO WHILE INKEY$ = ""
  LOOP
  END
END IF

END SUB

SUB extract (AInFile AS STRING, AOutFile AS STRING, AOffset AS LONG, ASize AS LONG)
DIM ffi  AS INTEGER 'free  infile number
DIM ffo  AS INTEGER 'free outfile number
DIM size AS LONG    'size of data to be extracted

CONST constbuffersize = 1024
DIM constbuffer AS STRING * constbuffersize 'buffer of constant size
DIM buffer      AS STRING                   'buffer of variable size

ffi = FREEFILE
OPEN AInFile FOR BINARY AS ffi
ffo = FREEFILE
OPEN AOutFile FOR BINARY AS ffo

SEEK ffi, AOffset + 1

size = ASize
DO WHILE size >= constbuffersize
  GET ffi, , constbuffer
  PUT ffo, , constbuffer
  size = size - constbuffersize
LOOP
    
buffer = SPACE$(size)
GET ffi, , buffer
PUT ffo, , buffer

CLOSE ffi, ffo
PRINT AOutFile + " extracted."
END SUB

FUNCTION fileExists% (AFilename AS STRING)

DIM ff AS INTEGER 'free file number
DIM fs AS LONG    'filesize

ff = FREEFILE
OPEN AFilename FOR BINARY AS ff
fs = LOF(ff)
CLOSE ff

IF fs = 0 THEN
  KILL AFilename
  fileExists% = FALSE
ELSE
  fileExists% = TRUE
END IF

END FUNCTION

SUB fillTable (ATable() AS tableType)
DIM ffi AS INTEGER 'free file number
DIM N   AS INTEGER 'number of files in archive

ffi = FREEFILE
OPEN ARCHIVEFILE FOR BINARY AS ffi
GET ffi, , N
REDIM ATable(0 TO N - 1) AS tableType
DIM hashTable(0 TO N - 1) AS INTEGER


'fill offset column in ATable
'and fill hashTable
FOR i% = 0 TO N - 1
  GET ffi, , hashTable(i%)
  GET ffi, , ATable(i%).offset
NEXT i%


'fill filesize column in ATable
FOR i% = 0 TO N - 2
  ATable(i%).filesize = ATable(i% + 1).offset - ATable(i%).offset
NEXT i%
ATable(N - 1).filesize = LOF(ffi) - ATable(N - 1).offset
CLOSE ffi


'fill filename column in ATable
'initialize
FOR i% = 0 TO N - 1
  ATable(i%).filename = LTRIM$(STR$(i%))
NEXT i%

CONST FILENAMESFILE = "FILENAME.TXT"
DIM f AS STRING 'filename
DIM index AS INTEGER

'create FILENAMESFILE if it doesn't exist already
OPEN FILENAMESFILE FOR BINARY AS ffi
CLOSE ffi

OPEN FILENAMESFILE FOR INPUT AS ffi
DO WHILE NOT EOF(ffi)
  INPUT #ffi, f
  index = search%(hash%(f), hashTable())
  IF index <> N THEN ATable(index).filename = f
LOOP
CLOSE ffi

END SUB

FUNCTION hash% (AFilename AS STRING)

f$ = UCASE$(AFilename)
dx% = 0
FOR i% = 1 TO LEN(f$)
  ax% = ASC(MID$(f$, i%, 1)) * 256
  dx% = (dx% XOR ax%)
  FOR j% = 1 TO 8
    bx% = dx%
    edx& = dx% * 2&
    IF &H7FFF < edx& THEN
      dx% = edx& - &H10000
    ELSEIF edx& < &H8000 THEN
      dx% = edx& + &H10000
    ELSE
      dx% = edx&
    END IF
    IF bx% < 0 THEN dx% = dx% XOR &H1021
  NEXT j%
NEXT i%

hash% = dx%

END FUNCTION

FUNCTION search% (AKey AS INTEGER, AArray() AS INTEGER)
'Bounded Linear Search
DIM x AS INTEGER
DIM y AS INTEGER

x = LBOUND(AArray)
y = UBOUND(AArray) + 1
DO WHILE x <> y
  IF AKey = AArray(x) THEN y = x ELSE x = x + 1
LOOP
search% = x

END FUNCTION

